# Name:         NHDPlusNavDS.py - NHDPlus Navigation Discovery Script
# Purpose:      Navigate downstream in the NHDPlusV2 stream network
#               and link starting points to user supplied objects
# Author:       R. Deffenbaugh - Horizon Systems Corporation
# Created:      11/10/2016
# Version       1.0.0.1
# Environment:  ArcGIS 10.2.2 and 10.5.1, Python 2.7 arcpy
# PythonWin 2.7.5 (default, May 15 2013, 22:43:36) [MSC v.1500 32 bit (Intel)] on win32.
# -------------------------------------------------------------------

# import the required libraries
import sys, os, time, string, win32com.client, shutil, datetime, csv, arcpy
from arcpy import env

## Beginning of user defined variables ##
#########################################

# Declare the name of the SQL Server database that will be used, within this script,
# by the NHDPlusV2VAANavigator to store the NHDPlus data and navigation results
DBname = "DSNavDBs"

# Declare the path to store the NHDPlusV2VAANavigator SQL Server database
# used to process the navigation
DBlocation = "D:\NHDPlusV21\NavigatorDBs"

# Declare the name of the comma delimited text file containing navigation starting points.
# This file has four fields or columns.
# 1. POIID, a user-assigned start point identifier
# 2. ComID of the flowline that associated with the POIID
# 3. StartMeasure which is the measure from which to start the navigation (Note: use -1 if when no StartMeasure is available).
# 4. MaxDistance to navigate (use 0 to navigate maximum distance)
# For example: 
# PariaRiver14_1,18267735,-1,0
# CO_River14_1,3528925,-1,0
NavStartPointsFN = 'D:\NHDPlusNavigation_DiscoveryScript\Input\NavStartPoints.csv'

# Declare the name of the comma delimited text file containing the paths to the NHDPlus data needed for navigations in this run.
# For example: 
# D:\NHDPlusV21\NHDPlusCO\NHDPlus14
# D:\NHDPlusV21\NHDPlusCO\NHDPlus15
NHDPlusDataFN = 'D:\NHDPlusNavigation_DiscoveryScript\Input\NHDPlusDataPaths.csv'

# Declare the name of the comma delimited text file containing the LinkObject IDs.
# This file contains objects that the user has linked to the NHDPlus network.
# This file has three fields or columns.
# 1. LinkObjID, a user-assign object identifier
# 2. ComID of the flowline that is linked to the Object
# 3. Measure, a location along the flowline.
# Note: This file must have a header line describing the fields in the file: LinkObjID,ComID,Measure
# For example: 
# LinkObjID,ComID,Measure
# LinkID14_1,18267741,0
# LinkID14_2,3528401,0
# LinkID15_3,20733853,0
# LinkID15_4,20733875,0
LinkObjectsFN = 'D:\NHDPlusNavigation_DiscoveryScript\Input\LinkObjects.csv'

# Declare the name of the output folder that will contain the output DBF file. This folder must exist.
OutputFolder = 'D:\NHDPlusNavigation_DiscoveryScript\Output'

# Declare a name for the output DBF file.  This file will contain the POIID, LinkObjID, Distance
# Note: This file will be overwritten if it exists
# Note: The distance field is intended to be the distance in KM between the bottom of the ComID holding the POIID and the bottom of the 
# ComID holding the LinkObjID.  In the vast majority of cases, the distance is very accurate.  However, due to the non-dendritic aspects 
# of the NHDPlusV2 network and the technique used to report results by the VAA navigator, the distance field (calculated from the 
# navigation results) is sometimes not as accurate as it could be.  Specifically, If the path between the POIID and the LinkOBJID includes 
# one or more minor paths downstream of network divergences, the distance field may only be an approximation of the actual
# distance.  On rare occasions, the distance field can be quite incorrect.  These cases can be identified where the POIID TerminalPathID 
# is not equal to the LINKOBJID TerminalPathID. 
OutputFileName = 'POILNKDST.DBF'

# Declare the name of the working folder for data processing. 
# This folder may or may not exist and will be created if it doesn't.
WorkingFolder =  "D:\Working"

## End of user defined variables ##
###################################

# Initialize the NHDPlusV2VAANavigator objects
o1 = win32com.client.Dispatch("NHDPlusV2VAANavigator.clsLoadSQLServerDB")
o2 = win32com.client.Dispatch("NHDPlusV2VAANavigator.clsMakeWorkingTable")
o3 = win32com.client.Dispatch("NHDPlusV2VAANavigator.clsV2Navigator")

# Create a session_id for this run which is added to the working folder.  Each execution of the script has a unique session id.
now = datetime.datetime.now()
session_id = datetime.datetime.strftime(now, "%Y%m%d%H%M%S%f")[:16]

# Create the temporary working folder
TmpWorkAreaPath = WorkingFolder + "\\" + session_id
if not os.path.exists(TmpWorkAreaPath):
    os.makedirs(TmpWorkAreaPath)

# Declare the variable used to increment the OID field
# in file named in the variable: OutputFileName
OIDincrement = 0

# Set the boolean variable that determines how to create 
# the SQL Server database named in the variable: DBname
# and how to copy other tables to the working folder
AddToExist = False

# Load the SQL Server database (DBname) with the NHDPlus VPUs needed for the navigations in this run.
# NHDPlusDataFN contains the paths to each VPU that needs to be loaded
print "Running NHDPlusNavDS.py"
print "Using the Temporary Working Folder: " + TmpWorkAreaPath + " to process data."
with open(NHDPlusDataFN) as getdatapath:
    reader = csv.reader(getdatapath)
    print "Running Module LoadSQLServerDB"
    try:
        for row in reader:
            # Retrieve the path to the NHDPlus data
            NHDPlusDataPath = row[0]
            print "Loading NHDPlus data from " + NHDPlusDataPath
            # Set the variable for the return value from LoadSQLServerDB
            intReturn = 0
            # Set common properties for LoadSQLServerDB
            o1.DatabaseLocation = DBlocation
            o1.DatabaseName = DBname
            o1.TempWorkAreaPath = TmpWorkAreaPath
            o1.SQLConnectionTimeout = 120
            o1.SQLCommandTimeout = 120
            o1.AddToExisting = AddToExist
            o1.InputNHDPlusLocation = NHDPlusDataPath
            #Call the method to load the data for the VPU
            intReturn = o1.LoadSQLServerDB
            
            # Check for any errors returned from LoadSQLServerDB
            if intReturn > 0:
               print "Return Value: " + str(intReturn)
               print "ProcessStatus: " + str(o1.ProcessStatus)
               print "ProcessMessage: " + o1.ProcessMessage
               exit()

            # After the first VPU's data is loaded into the SQL Server database
            # named in the variable DBname, add any additional VPUs.
            if not AddToExist:
                # Display the name of the SQL Server localDB instance used by the NHDPlus Navigation Discovery Script
                # This instance is used to log into SQL Server Management Studio to see the database named in the variable: DBname
                print "SQL Server localDB instance used by LoadSQLServerDB: " + o1.SQLDataSource
                
                # Copy the PlusFlowlineVAA table from the first VPU into the PlusFlowlineVAA table in the working folder
                arcpy.CopyRows_management(NHDPlusDataPath + "\\NHDPlusAttributes\\PlusFlowlineVAA.DBF", TmpWorkAreaPath + "\\PlusFlowlineVAA.DBF")
                # Change AddToExist so any additional VPUs will be added
                AddToExist = True
            else:
                # otherwise, append a copy of the current VPU's PlusFlowlineVAA table to the one already in the working folder
                arcpy.Append_management(NHDPlusDataPath + "\\NHDPlusAttributes\\PlusFlowlineVAA.DBF", TmpWorkAreaPath + "\\PlusFlowlineVAA.DBF")
           
        #End loop to load the SQL Server database (DBname) with the NHDPlus VPUs data needed to navigate

    except csv.Error as err:
        sys.exit('file %s, line %d: %s' % (NHDPlusDataFN, reader.line_num, err))

# Delete any old output DBF so that the results of this run aren't appended to it
if os.path.isfile(OutputFolder + "\\" + OutputFileName):
    arcpy.Delete_management(OutputFolder + "\\" + OutputFileName)
    
# Create a fresh output DBF from the file name specified in the variable: OutputFileName
arcpy.CreateTable_management(OutputFolder,OutputFileName)
# Add fields
arcpy.AddField_management(OutputFolder + "\\" + OutputFileName, "POIID", "TEXT")
arcpy.AddField_management(OutputFolder + "\\" + OutputFileName, "LinkObjID", "TEXT")
arcpy.AddField_management(OutputFolder + "\\" + OutputFileName, "Distance", "DOUBLE")
# Get rid of the default field that CreateTable_management creates
arcpy.DeleteField_management(OutputFolder + "\\" + OutputFileName, "Field1")

## Start of main loop ##
# Open the file containing navigation starting points
# and loop through the records processing each starting point
print "Running Module MakeWorkingTable and Module VAANavigate"
with open(NavStartPointsFN) as nsp:
    reader = csv.reader(nsp)
    try:
        for row in reader:
            # Read the fields from NavStartPointsFN
            POIID = row[0]
            ComID = row[1]
            Measure = row[2]
            MaxDist = row[3]
            print "Processing Starting Point: " + POIID + " " + ComID + " " + Measure  + " " + MaxDist
            
            # For each navigation, call MakeWorkingTable followed by VAANavigate
            ## MakeWorkingTable ##
            # Set the variable for the return value from MakeWorkingTable
            intReturn = 0
            # Set common properties for MakeWorkingTable
            o2.DatabaseLocation = DBlocation
            o2.DatabaseName = DBname
            o2.TempWorkAreaPath = TmpWorkAreaPath
            o2.SessionID = session_id
            o2.SQLConnectionTimeout = 120
            o2.SQLCommandTimeout = 120
            o2.StartComid = ComID
            o2.Navtype = "DNDIV"
            intReturn = o2.MakeWorkingTable

            # Check for any errors returned from MakeWorkingTable
            if intReturn > 0:
                print "MakeWorkingTable Return Value: " + str(intReturn)
                print "MakeWorkingTable ProcessStatus: " + str(o2.ProcessStatus)
                print "MakeWorkingTable ProcessMessage: " + o2.ProcessMessage
                exit()
               
            ## VAANavigate ##
            # Set the variable for the return value from VAANavigate
            intReturn = 0
            # Set common properties for VAANavigate
            o3.DatabaseLocation = DBlocation
            o3.DatabaseName = DBname
            o3.TempWorkAreaPath = TmpWorkAreaPath
            o3.SessionID = session_id
            o3.SQLConnectionTimeout = 120
            o3.SQLCommandTimeout = 120
            o3.WorkingTableName =o2.WorkingTableName
            o3.StartComid = ComID
            o3.StartMeasure = Measure
            o3.NavType = "DNDIV"
            o3.MaxDistance = MaxDist
            o3.SaveResults = False
            # Output the navigation results to a text file
            o3.NavResultFile = TmpWorkAreaPath + "\\" + "NAVRESU.txt"
            intReturn = o3.VAANavigate

            # Check for any errors returned from VAANavigate
            if intReturn > 0:
                print "VAANavigate Return Value: " + str(intReturn)
                print "VAANavigate ProcessStatus: " + str(o3.ProcessStatus)
                print "VAANavigate ProcessMessage: " + o3.ProcessMessage
                exit()
           
            # Import the navigation results text file (o3.NavResultFile)into a DBF
            # Copy NAVRESU.txt to NAVRESU.DBF
            arcpy.CopyRows_management(TmpWorkAreaPath + "\\" + "NAVRESU.txt", TmpWorkAreaPath + "\\NAVRESU.DBF")

            # Turn the DBF file into a table view 
            arcpy.MakeTableView_management(TmpWorkAreaPath + "\\" + "NAVRESU.DBF", "NAVRESU")

            # Create LinkObjects.DBF from the file name specified in the variable: LinkObjectsFN
            arcpy.CopyRows_management(LinkObjectsFN, TmpWorkAreaPath + "\\LinkObjects.DBF")

            # Don't add the table names of NAVRESU and PlusFlowlineVAA to the fields when joining NAVRESU to LinkObjects.DBF
            # Note: Setting env.qualifiedFieldNames = False doesn't work for Table Views.
            # You have to qualify the fields with tablename.fieldname
            env.workspace = TmpWorkAreaPath
            env.qualifiedFieldNames = False
            env.overwriteOutput = True

            # Join the working PlusFlowlineVAA.DBF to the navigation results table view
            arcpy.AddJoin_management("NAVRESU","COMID",TmpWorkAreaPath + "\\PlusFlowlineVAA.DBF","COMID","KEEP_COMMON")
            with arcpy.da.SearchCursor("NAVRESU",['NAVRESU.COMID', 'PlusFlowlineVAA.PathLength'],"NAVRESU.COMID = " + ComID)  as cursor:
                for row in cursor:
                    StartingPOID_PathLength = row[1]

            # Join navigation results table view to LinkObjects.DBF
            arcpy.AddJoin_management("NAVRESU","COMID",TmpWorkAreaPath + "\\LinkObjects.DBF","COMID","KEEP_COMMON")

            # Get the objects that were linked to the navigation of the starting point
            arcpy.CopyRows_management("NAVRESU", TmpWorkAreaPath + "\\NavVAALnk.DBF")
            with arcpy.da.SearchCursor(TmpWorkAreaPath + "\\NavVAALnk.DBF",['LinkObjID','COMID', 'PathLength']) as cursor:
                for row in cursor:
                    LinkObjID = str(row[0])
                    LinkObjID_PathLength = row[2]
                    Distance = StartingPOID_PathLength - LinkObjID_PathLength
            
                    print "NHDPlusNavDS navigated downstream from " + POIID +  " and found link(s) to " + LinkObjID  + " at a distance of " + str(Distance)
                    with arcpy.da.InsertCursor(OutputFolder + "\\" + OutputFileName, ['OID','POIID','LinkObjID','Distance']) as cursor:
                        cursor.insertRow([OIDincrement,POIID,LinkObjID,Distance])
                        OIDincrement += 1
            ## End of main loop ##
            
    except csv.Error as err:
        sys.exit('file %s, line %d: %s' % (NavStartPointsFN, reader.line_num, err))

# Remove the joins and close the DBFs
arcpy.RemoveJoin_management("NAVRESU")
arcpy.Delete_management("NAVRESU")

# Empty and delete the TmpWorkAreaPath folder
shutil.rmtree(TmpWorkAreaPath) 

# Terminate the objects and remove any remaining locks
cursor = None
o1 = None
o2 = None
o3 = None

print " "
print "NHDPlus Navigation Discovery Script Completed Successfully!"
print "The output is stored in " + OutputFolder + "\\" + OutputFileName